/*
 * onyx-the-black-cat.c
 * Fractal Guru <reverse@put.as> http://reverse.put.as
 *
 * For use with Mac OS X Tiger v10.4.11
 *
 * Based on original code by Landon J. Fuller <landonf@opendarwin.org>
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define VERSION 0.1

// we need this to use the kinfo_proc structure for sysctl
#define __APPLE_API_UNSTABLE
#define SYSCTL_DEF_ENABLED
#define PROC_DEF_ENABLED

#include <mach/mach_types.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <stdbool.h> 
#include <sys/types.h> 

#include <sys/proc.h>
#include <sys/sysctl.h>

#include <stdint.h>
#include "structures.h"

static struct sysent *_sysent;
extern int nsysent;

typedef int	ptrace_func_t (struct proc *, struct ptrace_args *, int *);
typedef int	open_func_t   (struct proc *, struct open_args *, int *);
typedef int sysctl_func_t (struct proc *, struct sysctl_args *, int *);
typedef int chmod_func_t  (struct proc *, struct chmod_args *, int *);

static ptrace_func_t *real_ptrace;
static open_func_t *real_open;
static sysctl_func_t *real_sysctl;
static chmod_func_t *real_chmod;

static int our_ptrace (struct proc *p, struct ptrace_args *uap, int *retval)
{
	if (uap->req == PT_DENY_ATTACH) {
		printf("[onyx-the-black-cat] Blocking PT_DENY_ATTACH for pid %d.\n", uap->pid);
		return (0);
	} else {
		return real_ptrace(p, uap, retval);
	}
}

// Uncomment this to redirect the open syscall

static int our_open (struct proc *p, struct open_args *uap, int *retval)
{
 int error;
 size_t dummy = 0;
 char savedpath[MAXPATHLEN];
// copy from user space to kernel space
//    copyinstr()   Copies a NUL-terminated string, at most len bytes long, from user-space address uaddr to
//    kernel-space address kaddr.  The number of bytes actually copied, including the terminating 
//	  NUL, is returned in *done.
//	  int copyinstr(const void *uaddr, void *kaddr, size_t len, size_t *done);
	error = copyinstr(uap->path, (void *)savedpath, MAXPATHLEN, &dummy);
    if (!error)
        printf("open(%s, %x, %x)\n", savedpath, uap->flags, uap->mode);
    else
        printf("open(?, %x, %x)\n", uap->flags, uap->mode);

    return real_open(p, uap, retval);
}


/*
The system call function implements the system call. Its function prototype is 
defined in the <sys/sysent.h> header as: 
typedef int32_t sy_call_t(struct proc *, void *, int *);
where struct proc * points to the currently calling process, and void * 
points to the system call’s arguments’ structure, if there is any, and
int * to the return value. 
*/
// bsd/vfs/vfs_syscalls.c
/* find the function that implements the syscal at bsd/kern/init_sysent.c */

static int our_chmod (struct proc *p, struct chmod_args *uap, int *retval)
{
 int error;
 size_t dummy = 0;
 char savedpath[MAXPATHLEN];
 //      chmod(const char *path, mode_t mode);
 error = copyinstr(uap->path, savedpath, MAXPATHLEN, &dummy);
 if (!error)
	printf("chmod requested to filename: %s\n",savedpath);
 else
	printf("chmod requested to unknown filename\n");
 // pass control to the real chmod function
 return real_chmod(p, uap, retval);
}

// our sysctl so we can intercept the anti-debug call
/*
 function that implements the sysctl syscall
         {AC(__sysctl_args), _SYSCALL_CANCEL_NONE, KERNEL_FUNNEL, (sy_call_t *)__sysctl, munge_wwwwww, munge_dddddd, _SYSCALL_RET_INT_T}, // 202 = __sysctl
 so it's "__sysctl"
*/
// __sysctl(struct proc *p, struct __sysctl_args *uap, __unused register_t *retval)
// retval is unused
// sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp, size_t newlen);

static int our_sysctl (struct proc *p, struct sysctl_args *uap, int *retval)
{
	int mib[4];
	int result;
	struct kinfo_proc kpr;
	
	// call the real_sysctl function and hold the result so we can use it to return from our hijacked sysctl
	result = real_sysctl(p,uap,retval);
 
 /* 
 **************************************************
  How to bypass the anti-debug protection:
  If we copy the kinfo_proc structure to kernel space and edit the p_flag and copy back to userspace,
  then we can bypass the anti-debug protection 
 ***************************************************
 */
	// int copyin(const void *uaddr, void *kaddr, size_t len); 
	// copyin() Copies len bytes of data from the user-space address uaddr to the kernel-space address kaddr.
	// copy structure from userspace to kernel space so we can verify if it's what we are looking for
	copyin(uap->name, &mib, sizeof(mib));
	// if it's a anti-debug call
	if (mib[0]==CTL_KERN && mib[1]==KERN_PROC && mib[2]==KERN_PROC_PID) 
	{
		printf("[onyx-the-black-cat] Detected possible sysctl anti-debug trick! Patching...\n");
		// then copy the result from the destination buffer ( *oldp from sysctl call) to kernel space so we can edit
		copyin(uap->old,&kpr,sizeof(kpr)); 
		// modify the p_flag because:
		// We're being debugged if the P_TRACED flag is set.
		// return ( (info.kp_proc.p_flag & P_TRACED) != 0 );
		kpr.kp_proc.p_flag = 0x0;
		// copy back to user space the modified structure
		// int copyout(const void *kaddr, void *uaddr, size_t len);
		// copyout() Copies len bytes of data from the kernel-space address kaddr to the user-space address uaddr.
		copyout(&kpr, uap->old,sizeof(kpr));
   /* sysctl()
     If the old value is not desired, oldp and
     oldlenp should be set to NULL.
   */
 }
 // and return from our hijacked function
 return result;
}


/*
 * nsysent is placed directly before the hidden sysent, so skip ahead
 * and sanity check that we've found the sysent array.
 *
 * Clearly, this is extremely fragile and not for general consumption.
 */
static struct sysent *find_sysent () 
{
	unsigned int table_size;
	struct sysent *table;

	table_size = sizeof(struct sysent) * nsysent;
	table = (struct sysent *) ( ((char *) &nsysent) + sizeof(nsysent) );

#if __i386__
	/* For reasons unknown, the table is offset by an additional 28 bytes on my i386 system */
	table = (struct sysent *) ( ((uint8_t *) table) + 28);
#endif

	printf("[onyx-the-black-cat] Found nsysent at %p (count %d), calculated sysent location at %p.\n", &nsysent, nsysent, table);

	/* Sanity check */
	printf("[onyx-the-black-cat] Sanity check: exit():%x fork():%x read():%x wait4():%x ptrace():%x , ",
		table[SYS_exit].sy_narg,
		table[SYS_fork].sy_narg,
		table[SYS_read].sy_narg,
		table[SYS_wait4].sy_narg,
		table[SYS_ptrace].sy_narg);

	if (
		table[SYS_exit].sy_narg == 1  &&
		table[SYS_fork].sy_narg == 0 &&
		table[SYS_read].sy_narg == 3 &&
		table[SYS_wait4].sy_narg == 4 &&
		table[SYS_ptrace].sy_narg == 4)
	{
		printf("sysent sanity check succeeded.\n");
		return table;
	} else {
		printf("sanity check failed, could not find sysent table.\n");
		return NULL;
	}
}

// THE FUN STARTS HERE
kern_return_t onyx_the_black_cat_start (kmod_info_t * ki, void * d) {
	printf("[onyx-the-black-cat] Starting patching ...\n");
    printf("[onyx-the-black-cat] Finding sysent table...\n");
	// get the sysent address
	_sysent = find_sysent();
	if (_sysent == NULL) {
	    printf("[onyx-the-black-cat] Error: Cannot find sysent table\n");
		return KERN_FAILURE;
	}
	
	// save address for the real functions
	real_ptrace = (ptrace_func_t *) _sysent[SYS_ptrace].sy_call;
	real_sysctl = (sysctl_func_t *) _sysent[SYS___sysctl].sy_call;
//	real_chmod = (chmod_func_t *)_sysent[SYS_chmod].sy_call;
//	real_open = (open_func_t *) _sysent[SYS_open].sy_call;
	
	// replace the original functions with ours, aka HIJACK
	_sysent[SYS_ptrace].sy_call = (sy_call_t *) our_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) our_sysctl;
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) our_chmod;
//	_sysent[SYS_open].sy_call = (sy_call_t *) our_open;

	printf("[onyx-the-black-cat] Patching is successful! Have fun...\n");
	// ALL DONE
	return KERN_SUCCESS;
}


// THE FUN ENDS HERE
kern_return_t onyx_the_black_cat_stop (kmod_info_t * ki, void * d) {
	printf("[onyx-the-black-cat] Starting unpatching ...\n");
	_sysent[SYS_ptrace].sy_call = (sy_call_t *) real_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) real_sysctl;
//	_sysent[SYS_open].sy_call = (sy_call_t *) real_open;
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) real_chmod;
	printf("[onyx-the-black-cat] Unpatching successful ...\n");
    // ALL DONE
	return KERN_SUCCESS;
}
